from .BaseOperator import BaseOperator
# from connectors_to_databases.BaseOperator import BaseOperator
from urllib.parse import quote
from typing import Union, Iterable

from sqlalchemy import create_engine, engine


class PostgreSQL(BaseOperator):
    """
    Connector to PostgreSQL database
    """
    def __init__(
            self,
            host: str = 'localhost',
            port: int = 5432,
            database: str = 'postgres',
            login: str = 'postgres',
            password: str = 'postgres'
    ):
        """
        :param host: Host/IP database; default 'localhost'.
        :param database: name database; default 'localhost'.
        :param port: port database; default 5432.
        :param login: login to database; default 'postgres'.
        :param password: password to database; default 'postgres'.
        """
        super().__init__(host, port, database, login, password)
        self._host = host
        self._database = database
        self._login = login
        self._password = password
        self._port = port

    def _authorization_database(self) -> engine.base.Engine:
        """
        Creating connector engine to database PostgreSQL.
        """

        engine_str = f'postgresql://' \
                     f'{self._login}:{quote(self._password)}@{self._host}:{self._port}/' \
                     f'{self._database}'

        return create_engine(engine_str)

    @staticmethod
    def list_columns_in_str_with_double_quotes(list_columns: list = None) -> str: # noqa: RUF013
        """
        **Function: list_columns_in_str_with_double_quotes**

        This static function takes a list of columns as the `list_columns` parameter and returns a string where each
        column value is enclosed in double quotes.

        **Parameters:**
        - `list_columns` (list, optional): The list of columns to be enclosed in double quotes.
        If not specified, defaults to `None`.

        **Return:**
        - `str`: A string containing column values enclosed in double quotes and separated by commas.

        **Example Usage:**

        ```python
        columns = ['column1', 'column2', 'column3']
        result = MyClass.list_columns_in_str_with_double_quotes(columns)
        print(result)
        ```

        **Output:**

        ```
        "column1", "column2", "column3"
        ```

        In this example, we pass the `columns` list of columns to the `list_columns_in_str_with_double_quotes`
        function and store the result in the `result` variable. Then, we print the value of `result`, which will
        contain the strings from the `columns` list enclosed in double quotes and separated by commas.


        @param list_columns: The list of columns to be enclosed in double quotes. If not specified, defaults
            to `None`.; default 'None'
        @return: A string containing column values enclosed in double quotes and separated by commas.
        """

        return ', '.join([f"\"{value}\"" for value in list_columns])

    @classmethod
    def generate_on_conflict_sql_query(
            cls,
            source_table_schema_name: str = 'public',
            source_table_name: str = None, # noqa: RUF013
            target_table_schema_name: str = 'public',
            target_table_name: str = None, # noqa: RUF013
            list_columns: Iterable[str] = None, # noqa: RUF013
            pk: Union[str, list] = 'id',
            replace: bool = False,
    ) -> str:
        """
        **Function: generate_on_conflict_sql_query**

        This class method generates an SQL query for performing data insertion with conflict handling using a
        specified primary key.

        **Parameters:**
        - `source_table_schema_name` (str, optional): The schema name of the source table. Defaults to `'public'`.
        - `source_table_name` (str): The name of the source table from which data will be inserted.
        - `target_table_schema_name` (str, optional): The schema name of the target table where data will be inserted.
            Defaults to `'public'`.
        - `target_table_name` (str): The name of the target table where data will be inserted.
        - `list_columns` (Iterable[str], optional): An iterable object containing the names of columns to be inserted.
            If not specified, all columns from the source table will be inserted.
        - `pk` (Union[str, list]): The primary key for checking insertion conflicts. It can be a string representing
            a single column name or a list of column names.
        - `replace` (bool): A flag indicating whether to replace existing data in case of conflicts. Defaults to
            `False`, which means conflicts will be ignored and nothing will be done.

        **Return:**
        - `str`: The generated SQL query for data insertion with conflict handling.

        **Example Usage:**

        ```python
        source_table_name = 'source_table'
        target_table_name = 'target_table'
        columns = ['column1', 'column2', 'column3']
        sql_query = MyClass.generate_on_conflict_sql_query(
            source_table_name=source_table_name,
            target_table_name=target_table_name,
            list_columns=columns,
            pk='id',
            replace=True
        )
        print(sql_query)
        ```

        **Output:**

        ```
        INSERT INTO public.target_table
        (
            "column1", "column2", "column3"
        )
        SELECT
            "column1", "column2", "column3"
        FROM
            public.source_table
        ON CONFLICT ("id") DO UPDATE SET
            "column1" = EXCLUDED."column1",
            "column2" = EXCLUDED."column2",
            "column3" = EXCLUDED."column3"
        ```

        In this example, we pass the source table name (`source_table_name`), target table name (`target_table_name`),
        column list (`columns`), primary key (`pk`), and the `replace` flag to the `generate_on_conflict_sql_query`
        function. Then, we store the result in the `sql_query` variable and print it. The output will contain the
        generated SQL query for data insertion with conflict handling.

        :param source_table_schema_name: The schema name of the source table; default `'public'`.
        :param source_table_name: The name of the source table from which data will be inserted; default `None`.
        :param target_table_schema_name: The schema name of the target table where data will be
            inserted; default `'public'`.
        :param target_table_name: The name of the target table where data will be inserted; default `None`.
        :param list_columns: An iterable object containing the names of columns to be inserted; default `None`.
        :param pk: The primary key for checking insertion conflicts. It can be a string representing a single
        column name or a list of column names; default `'id'`.
        :param replace: A flag indicating whether to replace existing data in case of conflicts.
        Defaults to `False`, which means conflicts will be ignored and nothing will be done; default `False`.
        :return: The generated SQL query for data insertion with conflict handling.
        """

        pk = cls.list_columns_in_str_with_double_quotes(list_columns=pk) if isinstance(pk, list) else f'"{pk}"'

        if replace:
            replace = f'''DO UPDATE SET {', '.join([f'"{i}" = EXCLUDED."{i}"' for i in list_columns])}'''
        else:
            replace = 'DO NOTHING'


        sql = f'''
        INSERT INTO {target_table_schema_name}.{target_table_name} 
        (
            {cls.list_columns_in_str_with_double_quotes(list_columns=list_columns)}
        )
        SELECT 
            {cls.list_columns_in_str_with_double_quotes(list_columns=list_columns)} 
        FROM 
            {source_table_schema_name}.{source_table_name}
        ON CONFLICT ({pk}) {replace}
        '''

        return sql # noqa: RET504



from PostgreSQL import PostgreSQL

pg = PostgreSQL(port=1)

# df = pg.execute_to_df('select 1 as one')

# check = pg.check_connection_to_database()

# print(check)

# pg.generate_on_conflict_sql_query(
#     # source_table_schema_name='stg',
#     source_table_name='tmp_test_pk',
#     # target_table_schema_name='dds',
#     target_table_name='test_pk',
#     list_columns=['id', 'value'],
#     # pk=['id','username'],
#     replace=True
# )

t = pg.check_connection_to_database()

print(t)



"""
:authors: k0rsakov
:license: Apache License, Version 2.0, see LICENSE file
:copyright: (c) 2022 k0rsakov
"""


# __author__ = 'k0rsakov'
# __version__ = '1.0.0'
# __email__ = 'korsakov.iyu@gmail.com'


from urllib.parse import quote

from sqlalchemy import create_engine, engine

from .BaseOperator import BaseOperator

# TODO: Change sqlalchemy to clickhouse-driver

class ClickHouse(BaseOperator):
    """
    Connector to PostgreSQL database
    """
    def __init__(
            self,
            host: str = 'localhost',
            port: int = 8123,
            login: str = 'default',
            password: str = 'default'
    ):
        """
        :param host: Host/IP database; default 'localhost'.
        :param port: port database; default '8123'.
        :param login: login to database; default 'default'.
        :param password: password to database; default 'default'.
        """
        super().__init__(host, port, login, password)
        self._host = host
        self._login = login
        self._password = password
        self._port = port

    def _authorization_database(self) -> engine.base.Engine:
        """
        Creating connector engine to database ClickHouse.
        """

        engine_str = f'clickhouse://{self._login}:{quote(self._password)}@{self._host}:{self._port}/default'

        return create_engine(engine_str)



from urllib.parse import quote
from typing import Union

from sqlalchemy import create_engine, engine

import pandas as pd

from .TypeHinting import SQLQuery


class BaseOperator:
    """
    BaseOperator for databases
    """
    def __init__(self,
                 host: str = 'localhost',
                 port: int = None,
                 database: str = None,
                 login: str = None,
                 password: str = None,
                 ):
        """
        :param host: Host/IP database; default 'localhost'.
        :param database: name database; default 'None'.
        :param port: port database; default 'None'.
        :param login: login to database; default 'None'.
        :param password: password to database; default 'None'.
        """
        self._host = host
        self._database = database
        self._login = login
        self._password = password
        self._port = port

    def _authorization_database(self) -> engine.base.Engine:
        """
        Creating connector engine to database PostgreSQL.
        """

        engine_str = f'base://' \
                     f'{self._login}:{quote(self._password)}@{self._host}:{self._port}/' \
                     f'{self._database}'

        return create_engine(engine_str)

    def insert_df(
            self,
            df: pd.DataFrame = None,
            table_name: str = None,
            table_schema: str = None,
            chunksize: Union[int, None] = 10024,
            index: bool = False,
            if_exists: str = 'append',
    ) -> Union[None, Exception]:
        """
        Inserting data from dataframe to database.

        :param df: dataframe with data; default None.
        :param table_name: name of table; default None.
        :param table_schema: name of schema; default None.
        :param chunksize: Specify the number of rows in each batch to be written at a time.
            By default, all rows will be written at once; default `10024`.
        :param if_exists: {'fail', 'replace', 'append'}, default 'append'
            How to behave if the table already exists.

            * fail: Raise a ValueError.
            * replace: Drop the table before inserting new values.
            * append: Insert new values to the existing table.
        :param index: Write DataFrame index as a column. Uses `index_label` as the column
            name in the table.
        """

        df.to_sql(
            name=table_name,
            schema=table_schema,
            con=self._authorization_database(),
            chunksize=chunksize,
            index=index,
            if_exists=if_exists,
        )

    def execute_to_df(
            self,
            sql_query: str = SQLQuery,
    ) -> Union[pd.DataFrame, Exception]:
        """
        Getting data from database with SQL-query.

        :param sql_query; default `''`.
        :return: DataFrame with data from database.
        """

        return pd.read_sql(
            sql=sql_query,
            con=self._authorization_database(),
        )

    def execute_script(
            self,
            manual_sql_script: SQLQuery
    ) -> None:
        """
        Execute manual scripts (INSERT, TRUNCATE, DROP, CREATE, etc). Other than SELECT

        :param manual_sql_script: query with manual script; default `''`.
        :return: None.
        """
        self._authorization_database().execute(manual_sql_script)

    def get_uri(self) -> engine.base.Engine:
        """
        Get connector for manual manipulation with connect to database.

        :return engine.base.Engine:
        """

        return self._authorization_database()

    def check_connection_to_database(self) -> Union[bool, Exception]:
        """
        Method to check connection to database.

        :return: boolean True, if connection to database is successful, Exception otherwise.
        """
        df = self.execute_to_df('SELECT 1 AS ONE')

        return bool(isinstance(df, pd.DataFrame))



from io import open
from setuptools import setup


version = '1.0.5'

with open('README.md', encoding='utf-8-sig') as f:
    long_description = f.read()

setup(
    name='connectors_to_databases',
    version=version,

    author='k0rsakov',
    author_email='korsakov.iyu@gmail.com',

    description=(
        u'Python module for connect with PostgreSQL and ClickHouse '
    ),
    long_description=long_description,
    long_description_content_type='text/markdown',

    url='https://github.com/k0rsakov/connectors_to_databases',
    download_url='https://github.com/k0rsakov/connectors_to_databases/archive/refs/heads/main.zip',

    license='Apache License, Version 2.0, see LICENSE file',
    
    # TODO: deprecated psycopg2, clickhouse-sqlalchemy
    packages=['connectors_to_databases'],
    install_requires=[
        'SQLAlchemy==1.4.44',
        'pandas==1.5.1',
        'psycopg2==2.9.5',
        'clickhouse-sqlalchemy==0.2.3',
        'clickhouse_driver==0.2.5',
        'ruff==0.0.278'
    ],

    classifiers=[
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Developers',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Programming Language :: Python :: Implementation :: CPython',
    ],
    
    python_requires='>=3.9',
    keywords=['database', 'SQL', 'PostgreSQL', 'ClickHouse', 'dataframe'],
)
